
/* ===============================================================
                       definition of constant
   =============================================================== */

#include <stdio.h>
#include <math.h>

#define TRUE  1
#define FALSE 0
#define ON    1
#define OFF   0


/*  Depending on Computer System: (Com)     
              Number of particles: m                 */
#define PARTICLE_NUM_MAX 50000

/* (Com) IMPORTANT: check your computer system         */
/* randum number   RAND_MAX is defined by  2^{8*4-1}   */
/* for (SUN, PC cluster) #define RAND_MAX 2147483647   */
/* for (HITACHI, SGI)    #define RAND_MAX 32767        */
#define RAND_MAX 32767

/*  Depending on Case Study: (Case)
              Maximum number of data: >= N           */
#define DATA_MAX 200

/* (Case) when you analyze an integer sequence, 
                 Y_TYPE_INT ON   */
#define Y_TYPE_INT OFF


/*            state space model                      
    xn[0] : t_n
    xn[1] : log10 sigma^2_n
    xn[2] : log10 tau^2                              */
#define STATE_DIM_MAX 3 

/*            system noise 
    vn[0] : epsilon_n                                */
#define SYSNOISE_DIM_MAX 1

/* number of lag for fixed lag smoothing: l          */
#define LAG_MAX 20

/* MPM ON          : Marginal Posterior Mode 
       OFF         : Marginal Posterior Average  


   SMOOTH_PAL_MAX  : maximum number of the parallel smoothing 

   PAL_SMO_AVE ON  : the result of parallel smoothing is defined 
                     by taking an average of each result of smooting 
               OFF : ... 
                     by the result with minimum AIC value 
   PRDFIL_PAL_MAX  : maximum number of the parallel procedure for 
                     getting many AIC values
*/
#define MPM ON
#define SMOOTH_PAL_MAX 5

#define PAL_SMO_AVE  OFF
#define PRDFIL_PAL_MAX 100

#define Gauss_dF_DIM_MAX 10000
#define Gauss_dF_range   4.0

static double  xmin_gpdf, xmax_gpdf, dx_gpdf, dx_unif, 
       dF[Gauss_dF_DIM_MAX] ;
/* global variable specifying data point n */
static int LOC_n ;

/* initial distribution of state vector */
static double init_type[STATE_DIM_MAX][3]={  0.0,  0.0,  4.0, 
                                            -1.0, -4.0,  4.0,
                                            -1.0, -4.0,  2.0 } ;
/* ====> corresponds to 
             t_0       ~  Normal(0, 4)
       log10 sigma_0^2 ~  Uniform([-4, 4])
       log10 tau^2     ~  Uniform([-4, 2])   */

static double sys_type[SYSNOISE_DIM_MAX][3]={ 1.0, 0.0, 0.00001 } ;
/* ====> corresponds to 
             epsilon_n ~ Cauchy(0,0.00001)  
                         hyper-hyper-parameter d^2=0.00001  */

static double  obs_type[3]={0.0, 0.0, 1.0} ;
/* ====> corresponds to  N(0,1) */


/*            internal expression */
#define BOOL int
#define US unsigned short 
#define US_BIT 15

#define GA_ONOFF OFF
#define INTER_MARRIAGE OFF
static double Pc[STATE_DIM_MAX] = {0.05} ;
static double Pm[STATE_DIM_MAX] = {0.05} ;
#define BitcMax 16
#define BitcMin  9
#define BitmMax 16
#define BitmMin  9


#define S_MAX 65535
#define S_MIN 0
/*  range of values of system noise variables: R_MIN <---> RMAX */
#define R_MAX  5.0
#define R_MIN -5.0
#define R2S (S_MAX - S_MIN)/(R_MAX - R_MIN)
#define S2R (R_MAX - R_MIN)/(S_MAX - S_MIN)

#define MY_OVERFLOW   1.0e+38
#define MY_UNDERFLOW  1.0e-38
#define MISSING -9
#define LOG_Y 0

#define NULL 0
#define EXIST_FILE 0

#define NA 0.0
#define PI  3.14159265358979312

#define PrintM -40
#define SHOW_MAX 3
/* number of output variables 
   See function decomposition() shown below                 */

#define ULout 3
/* ULout:1  --> output median for each variable of interest 
   ULout:3  -->  (median, -1 sigma, +1sigma)                */
static double ULV[ULout] = {0.5, 0.1587, 0.8413} ;

/* ((( output file ))
   (Number of data) x (1 + ULout x SHOW_MAX ) 
   
   ex. for ULout=3 and SHOW_MAX=2
   
   L1: y_1, median of var1_1, -1 sigma of var1_1, +1 sigma of var1_1, 
            median of var2_1, -1 sigma of var2_1, +1 sigma of var2_1
   L2: y_2, median of var1_2, -1 sigma of var1_2, +1 sigma of var1_2, 
            median of var2_2, -1 sigma of var2_2, +1 sigma of var2_2
                  ............................
                                                                */


 /* ===============================================================
                         state space model
    =============================================================== */

/*  [ for prediction ] */

void one_ahead_GA( xn )
     double xn[] ;
    {
     double xnext[STATE_DIM_MAX], t_phase ;
       
/* scaling for a variable of f_n */

         xnext[0] = 2.0*xn[0]-xn[1] ; 
         xnext[1] = xn[0]; 

            xn[0] = xnext[0] ;
            xn[1] = xnext[1] ;
            
     }


void one_ahead( xn, vn )
     double xn[], vn[] ;
    {
     double xnext[STATE_DIM_MAX], sv1, rcauchy(), sv_temp ;
     int    ii ;

     /* SYSTEM MODEL */
     
         /* t_n = t_{n-1} + v_n 
            sv1  <==> tau^2 (=dispersion parameter in Cauchy distribution */
             sv1 = exp( xn[2]*log(10.0) ) ;
         sv_temp = rcauchy(1.0, 0.0, sv1) ; 
         xnext[0] = xn[0] + sv_temp ;
         
         /* log10 sigma^2_n = log10 sigma^2_{n-1} + epsilon_n */
         xnext[1] = xn[1] +  vn[0] ;
         xnext[2] = xn[2] ;

         /* copy x_{n-1} to x_n */
         for( ii=0 ; ii <  STATE_DIM_MAX ; ++ii ) xn[ii] = xnext[ii] ; 
     }

double get_likelihood( yn, xn )
    double yn, xn[] ;
    {
    double err_f, en, bv, mu, tau2, gauss(), cauchy(), pearson(), 
           poisson(), binomial(), plam ;

    /* OBSERVATION MODEL */
    /* y_n = t_n + w_n  | w_n == err_f  */ 
    err_f = yn - xn[0] ;

    bv   = obs_type[0] ; mu = obs_type[1] ; 
    tau2 = obs_type[2] * exp( xn[1]*log(10.0) ) ;
    /* tau2 <==> sigma^2_n */


         if( bv == 0.0 )   en =  gauss(bv, mu, tau2, err_f) ; 
    else if( bv == 1.0 )   en = cauchy(bv, mu, tau2, err_f) ; 
    else if( bv == -10.0)  en = poisson(yn, plam) ;
    else if( bv == -20.0)  en = binomial(yn, plam) ;
    else                   en = pearson(bv, mu, tau2, err_f) ;

/*  for d w_n/d y_n ((( for OLD version )))
    en /= plam ;   */
    
    
    return( en ) ;
    }

double poisson( yn, pl )
    double yn, pl ;
    {
    double pdfva ;
    int    ix, imax, ii ;

    if( pl <= 0.0 ) pl = 1.0 ;
    pl = log( pl ) ;
    imax = ceil( yn ) ;

    pdfva = yn*pl - exp( pl ) ;
    for( ii=1 ; ii <= imax ; ++ii ) pdfva -= log( (double)ii ) ;
   
    pdfva = exp( pdfva ) ;
    return( pdfva ) ; 
    }

double binomial( yn, pl )
    double yn, pl ;
    {
     int ieve, ii ;
     double a3, pdfva ;

     a3 = log( 1.0 + exp( pl ) ) ;
     ieve = (int)yn ;
     
     if( ieve == 1 ) 
         pdfva = log( 2.0 ) + pl - 2.0*a3 ;
     else
         pdfva = yn*(pl - a3) - (2.0 - yn)*a3 ;
      
    pdfva = exp( pdfva ) ;
    return( pdfva ) ; 
    }

void decomposition( xn, dc )
double xn[], dc[] ;
{

     dc[0] =  xn[0] ;
     dc[1] =  xn[1] ;
     dc[2] =  xn[2] ;
}


void init_mutdep_vn( tpxn )
double tpxn[] ;

{
}

  
/* ------------------------------------------------------------------- */



