
#include "SOSSM.c"
#include "mcfbase.c"
#include "ssm.c"

/*   || READEME ||

PROGRAM CODE for Self-Organizing State Space Model (SOSSM)
   written by Tomoyuki Higuchi
              Institute of Statistical Mathematics (ISM)
              (email: higuchi@ism.ac.jp)
              (Web URL: http://www.ism.ac.jp/~higuchi/)

   References:
   
   1) Genshiro Kitagawa, 
   Self-organizing state space model, 
   Journal of the American Statistical Association, 
   Vol.93, No.443, 1203--1215, 1998.
   
   2) Tomoyuki Higuchi and Genshiro Kitagawa, 
   Knowledge discovery and Self-organizing state space model,
   IEICE Transaction on Information and Systems, 
   Vol. E83-D, No. 1, 36--43, 2000.
   
   NECESSARY C programs
   1) mainSOSSM.c (this program)
   2) SOSSM.c   ---- user defiend program code
   3) mcfbase.c ---- basic routines for Monte Carlo computation
   4) ssm.c     ---- basic routines for state space modeling
   
   Additional C program
   *) ckruni.c  ---- checking program for a random number generator
                     function in C, runi()

   COMPILE:
   %cc mainSOSS.c -lm -o SOSSM.out
   
   USAGE:
   %SOSSM.out input.dat output.dat 
          ex. input.dat   ==> SOSSMtest.dat 
              output.dat  ==> SOSSMout.dat  

   RECOMMENDATION:
   Range of data should be normalized to be around +-5. 

   ******************************************************
   When you apply this program to your analysis and make 
   your result(s) in public, please acknowledge me and 
   send me email (higuchi@ism.ac.jp) to report. 
   ******************************************************

*/

main(argc, argv)
    int argc;
    char *argv[];
    {
    double y[DATA_MAX], sy[(ULout*SHOW_MAX)][DATA_MAX] ;
    double parallel_smoothing(), tmp ;
    void  define_mcf_system(), dF_gauss(),  time_series_ascii_output(),
          read_data()   ;

    printf("  \n"); 
    printf(" PROGRAM: Self-Organizing State Space Model  (SOSSM) \n"); 
    printf("          Tomoyuki Higuchi (Inst. Stat. Math., 2002) \n"); 

    define_mcf_system()   ;  dF_gauss() ;
    read_data(argv[1], y)   ;

    tmp = parallel_smoothing( y, sy ) ; 
    time_series_ascii_output( y, sy, argv[2]) ; 
  
  }


