
/* 
         [   Monte Carlo Filter   ]

         designed by T. Higuchi I.S.M.
  
         latest Version Up data      95/Feb./06

*/

#include <stdio.h>
#include <math.h>

/* ===============================================================
                      definition of global variable
    =============================================================== */

static  int      N, K, M, L ;

/* ===============================================================
                      definition of struct
    ============================================================== */

/* particle */
struct particle {
       double state[STATE_DIM_MAX] ;
       double fitness ;
     } ;
typedef struct particle particle ;

/* population */
struct population {
       particle x[PARTICLE_NUM_MAX] ;
     } ;
typedef struct population population ;

/* family line */
struct family {
       particle history[LAG_MAX] ;
     } ;
typedef struct family family ;

 /* ===============================================================
                       Bit operations
    =============================================================== */

BOOL bit_tst(adr, bit)
     US *adr ;
     int bit;
    {
    US mask_bit = 0x0001 ;
  
    mask_bit <<= bit ;
    if( 0 != ((*adr)&mask_bit) ) return(TRUE ) ;
    else                         return(FALSE) ;
    }

void show_bit(adr)
     US *adr ;
{
     int i ;
                              printf("  |") ;
    for( i = US_BIT ; i > (-1) ; --i )
    {    
    if( TRUE == bit_tst(adr,i)) printf("1") ;
    else                        printf("0") ;
    if( (i%4) == 0 )            printf("|") ;
     }
    printf("   \n") ;     
}


void show_bit_flag(adr, flag)
     US *adr  ;
     int flag ;
{
     int i ;
                              printf("  |") ;
    for( i = US_BIT ; i > (-1) ; --i )
    {    
    if( TRUE == bit_tst(adr,i)) printf("1") ;
    else                        printf("0") ;
    if( (i%4) == 0 )            printf("|") ;
    if(  i == flag )           printf(":") ;
     }
    printf("   \n") ;     
}

void bit_flip(adr, bit)
     US *adr ;
     int bit ;
     {
     US mask_bit = 0x0001 ;
     mask_bit <<= bit ; *adr ^= mask_bit ;
     } 
    
void upper_bit_cut(adr, bit)
     US *adr ;
     int bit ;
   {
     US mask_bit = 0xFFFF ;
     mask_bit <<= bit ; *adr &= mask_bit ;
   }

void lower_bit_cut(adr, bit)
     US *adr ;
     int bit ;
   {
     US mask_bit = 0xFFFF ; 
     mask_bit >>= bit ; *adr &= mask_bit ;
   }

void crossover( adr1, adr2, bit)
     US *adr1, *adr2 ;
     int bit         ;
   {
     US up1, up2, low1, low2 ;
     int low_bit ;
     void lower_bit_cut(), upper_bit_cut() ;

     up1 = *adr1 ;  upper_bit_cut(&up1, bit) ;
     up2 = *adr2 ;  upper_bit_cut(&up2, bit) ;

    low_bit = US_BIT - bit + 1 ;
    low1 = *adr1 ;  lower_bit_cut(&low1, low_bit) ;
    low2 = *adr2 ;  lower_bit_cut(&low2, low_bit) ;

    *adr1 = up1 | low2 ;
    *adr2 = up2 | low1 ;
    }

void show_crossover(adr1, adr2, bit)
     US *adr1, *adr2 ;
     int bit ;
{
     void crossover(), show_bit_flag() ;

     printf(" \n");
     show_bit_flag(adr1, bit);  show_bit_flag(adr2, bit);
     crossover(adr1, adr2, bit) ; printf(" \n");
     show_bit_flag(adr1, bit);  show_bit_flag(adr2, bit);
}


 /* ===============================================================
                       encode and decode
    =============================================================== */

US encode( value, col_n )
    int col_n   ;
    double value ;

    {
     double tp ;
     if( value > R_MAX ) 
       { tp = R_MAX ; 
/*       printf(" Value exceeds upper limit (T.H.) %15.7f\n",value) ; */ } 
     else if( value < R_MIN ) 
       { tp = R_MIN ; 
/*       printf(" Value exceeds lower limit (T.H.) %15.7f \n",value) ; */ }
     else
         tp = (value - R_MIN)*R2S+S_MIN ;

     switch( col_n )
       {
        case 0  : return( (US)tp ) ; break ;
        default : return( (US)tp ) ; break ;
       }
     }

double decode( vus, col_n )
   int col_n   ;
   US  vus ;

    {
     double tp ;
     tp = ((double)(vus - S_MIN))*S2R+R_MIN ;
     switch( col_n )
       {
        case 0  : return( tp ) ; break ;
        default : return( tp ) ; break ;
       }
     }

void encode_state( x, indi )
        US *indi ;
     double *x ;
    {
    int i ;
    US  encode() ;

    for( i = 0 ; i < STATE_DIM_MAX ; ++i )
    *(indi +i) = encode( *(x+i), i ) ; 
    }  


void decode_state( indi, x )
     double *x ;
        US *indi ;
    {
    int i ;
    double  decode() ;

    for( i = 0 ; i < STATE_DIM_MAX ; ++i )
    *(x+i) = decode( *(indi+i), i ) ; 
    }  

 /* ===============================================================
                      operations of states
    =============================================================== */
void define_individual( indi )
     US *indi ;
    {
       *(indi+0) = 0x0000 ;
       *(indi+1) = 0x7FFF ;
    }  

void print_individual( indi )
     US *indi ;
    {
    int i ;
    for( i =0 ; i < STATE_DIM_MAX ; i++ )
    {  printf(" [%1d] ", i) ;
       show_bit((indi+i))   ; }
    printf("\n") ;
    }     
       


 /* ===============================================================
                  operations of population and states
    =============================================================== */

void trans_pop2state( pop, z, particle_num)
     population *pop ;
     double *z ;
     int   particle_num ;
    {
    int i1 ;
        for( i1 =0 ; i1 < STATE_DIM_MAX ; i1++ )
          *(z+i1) = (*pop).x[particle_num].state[i1] ;
    } 


void trans_state2pop( z, pop, particle_num)
     population *pop ;
     double *z ;
     int   particle_num ;
    {
    int i1 ;
        for( i1 =0 ; i1 < STATE_DIM_MAX ; i1++ )
          (*pop).x[particle_num].state[i1] = *(z+i1);
    }


int p_cross( ncp )
int ncp ;
{
     double pb, ss, runi() ;
     int    jc ;

     pb = runi() ;
         if( pb <= (double)(Pc[ncp]) ) {
             do {
                 ss = runi()*(double)( BitcMax - BitcMin ) ;
                 jc = (int)(ss) + BitcMin ;
                }
             while( jc == BitcMax ) ; }
         else     jc = 0 ;
     return( jc ) ;
}

int p_mute( ncp )
int ncp ;
{
     double pb, ss, runi() ;
     int    jc ;

     pb = runi() ;
     if( pb <= Pm[ncp] ) { 
         do {
              ss = runi()*(double)(BitmMax-(BitmMin-1)) ;
              jc = (int)(ss) + BitmMin ;
            }
         while( jc > BitmMax ) ; }
     else           jc = 0 ;
     return( jc ) ;
}

int pop_crossover( pop )
    population *pop ;
{
    int i, j, num_pair, mate1, mate2, 
        bit, p_cross(), occur, num_idpairs ;
    double x1[STATE_DIM_MAX], x2[STATE_DIM_MAX],  
           TpPc[STATE_DIM_MAX] ;

    /* stack Apr/19/1999 */
    static int    id[PARTICLE_NUM_MAX], ix[PARTICLE_NUM_MAX]  ;

    US     u1[STATE_DIM_MAX], u2[STATE_DIM_MAX], adr1, adr2 ;
    void rand_sort(), crossover(), trans_pop2state(),
         encode_state(), decode_state(), trans_state2pop() ;

    rand_sort(M, ix) ;

    if( INTER_MARRIAGE == 0 ) {
      /* get number of identical pairs */
    num_pair = M/2 ;    num_idpairs = 0 ;
    for( i=0; i < num_pair ; i++) 
    {
        mate1 = ix[(2*i)] ; mate2 = ix[(2*i+1)] ; 
        trans_pop2state( pop, x1, mate1 ) ;
        trans_pop2state( pop, x2, mate2 ) ;

        id[i] = (int)(STATE_DIM_MAX) ;
        for( j=0 ; j < STATE_DIM_MAX ; j++ )
        {   if( x1[j] == x2[j] ) id[i] -= 1 ; }

        if( id[i] == 0 ) num_idpairs++ ; 
     }

     for( j=0 ; j < STATE_DIM_MAX ; j++ ){
         TpPc[j] = Pc[j] ; 
         Pc[j]  *= 
          ( (double)(num_pair)/(double)(num_pair - num_idpairs) ) ;
         if( Pc[j] > 1.0 ) Pc[j] = 1.0 ;
     }
/*     printf(" identical pairs rate: %5.3f \n",
           (double)(num_idpairs)/(double)(num_pair) ) ; */
    occur = 0 ;
    for( i=0; i < num_pair ; i++) 
    {
        if( id[i] != 0 )
       {
        mate1 = ix[(2*i)] ; mate2 = ix[(2*i+1)] ; 
        trans_pop2state( pop, x1, mate1 ) ;
        trans_pop2state( pop, x2, mate2 ) ;
        encode_state( x1, u1 ) ; encode_state( x2, u2 ) ; 

        for( j=0 ; j < STATE_DIM_MAX ; j++ ){
        bit = p_cross( j ) ;
        if( bit > 0 )
            {
            occur++ ;
            adr1 = u1[j] ;  adr2 = u2[j] ;
            crossover(&adr1, &adr2, bit) ;
            u1[j] = adr1 ;  u2[j] = adr2 ;
            }
        }
        decode_state( u1, x1 ) ; decode_state( u2, x2 ) ; 
        trans_state2pop( x1, pop, mate1 ) ;
        trans_state2pop( x2, pop, mate2 ) ;
       }
    }

    for( j=0 ; j < STATE_DIM_MAX ; j++ )  Pc[j] = TpPc[j] ; 
    } /* intermarriage is prohibited   */
    else {

    occur = 0 ;
    num_pair = (int)(M)/2 ;
    for( i=0; i < num_pair ; i++) {
        mate1 = ix[(2*i)] ; mate2 = ix[(2*i+1)] ; 
        trans_pop2state( pop, x1, mate1 ) ;
        trans_pop2state( pop, x2, mate2 ) ;
        encode_state( x1, u1 ) ; encode_state( x2, u2 ) ; 

        for( j=0 ; j < STATE_DIM_MAX ; j++ )
        {
        bit = p_cross( j ) ;
        if( bit > 0 )
            {
            occur++ ;
            adr1 = u1[j] ;  adr2 = u2[j] ;
            crossover(&adr1, &adr2, bit) ;
            u1[j] = adr1 ;  u2[j] = adr2 ;
            }
        }
        decode_state( u1, x1 ) ; decode_state( u2, x2 ) ; 
        trans_state2pop( x1, pop, mate1 ) ;
        trans_state2pop( x2, pop, mate2 ) ;
/*
        if( bit > 0 )
        printf(" %10.5f %10.5f\n",x1[0]-tmp1,x2[0]-tmp2) ;
*/
    }

    } /* intermarriage is permitted */
    return( 2*occur );
}


int pop_mutation( pop )
    population *pop ;
{
    int i, j, bit, p_mute(), occur ;
    double  xn[STATE_DIM_MAX], tmp ;
    US     ind[STATE_DIM_MAX], adr ;
    void rand_sort(), bit_flip(), trans_pop2state(),
         encode_state(), decode_state(), trans_state2pop() ;

    occur = 0 ;
    for( i=0; i < M ; i++) {
        trans_pop2state( pop, xn, i ) ;
        encode_state( xn, ind ) ; 
        tmp = xn[0] ;

        for( j=0 ; j < STATE_DIM_MAX ; j++ ){

        bit = p_mute( j ) ;
        if( bit > 0 ) {
            occur++ ;
            adr = ind[j] ; 
            bit_flip(&adr, bit); 
            ind[j] = adr ;  }
        }
        decode_state( ind, xn ) ;
/*
        if( bit > 0 )
        printf(" %10.5f %10.5f %10.5f \n",tmp,xn[0],xn[0]-tmp) ;
        trans_state2pop( xn, pop, i ) ;
*/
    }
    return( occur ) ;
}


void pop_generation_shift( pop )
     population *pop ;
{
    int i ;
    double  xn[STATE_DIM_MAX] ;
    void    one_ahead_GA(), trans_pop2state(), trans_state2pop() ;

    for( i=0; i < M ; i++) {
       trans_pop2state( pop, xn, i ) ;
       one_ahead_GA( xn ) ;
       trans_state2pop( xn, pop, i ) ;
    }

}


 /* ===============================================================
                      operations of population
    =============================================================== */


void print_pop( pop )
     population *pop ;
    {
    int i1, i2 ;

    for( i1 = 0 ; i1 < M ; i1++ )
      {
        for( i2 =0 ; i2 < STATE_DIM_MAX ; i2++ )
        {  printf(" %10.5f", (*pop).x[i1].state[i2]) ;}
           printf("\n") ;
      }
    } 

/* Version UP 1995/11/28  */
void get_stat_pop( pop, st )
     population *pop  ;
     double      st[] ;
    {
    int is, ic, iu, mpm_sel[ULout]   ;
    void sort(), decomposition(), trans_pop2state() ;
    double xns[STATE_DIM_MAX], dcav, dc[SHOW_MAX] ;
    static double ra[PARTICLE_NUM_MAX] ;

    for( ic = 0 ; ic < (SHOW_MAX*ULout) ; ic++ ) st[ic] = 0.0 ;
    for( iu = 0 ; iu < ULout ; iu++ ) 
            mpm_sel[iu] = (int)( ((double)(M))*ULV[iu] ) ;

        for( ic = 0 ; ic < SHOW_MAX ; ic++ ) {
             dcav = 0.0 ;
             for( is = 0 ; is < M ; is++ ) {
                  trans_pop2state( pop, xns, is ) ;
                  decomposition( xns, dc ) ;
                  ra[is]  = dc[ic] ; dcav += dc[ic] ;  }
             sort( M, ra ) ;

             for( iu = 0 ; iu < ULout ; iu++ ) 
                 st[ ULout*ic + iu ] = ra[ mpm_sel[iu] ] ;

             if( MPM != ON ) st[ULout*ic] = dcav/((double)(M)) ;  
        }

    } 




void get_stat_2pop( pop1, pop2, st )
     population *pop1, *pop2  ;
     double      st[] ;
    {
    int is, ic  ;
    void sort() ;
    /* Apr/19/1999 */
    static double  ra[2*PARTICLE_NUM_MAX] ;

    for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ ) st[ic] = 0.0 ;

    if( MPM == 1 ) {

        for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ ) {
             for( is = 0 ; is < M ; is++ ) {
                  ra[is]      = (*pop1).x[is].state[ic] ;  
                  ra[is + M]  = (*pop2).x[is].state[ic] ;  }
             sort( 2*M, ra ) ;
             st[ic] = (ra[ M-1 ]+ra[ M ])/2.0 ;
        }
    }
    else {
        for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
        {
        for( is = 0 ; is < M ; is++ ) {
              st[ic] += (*pop1).x[is].state[ic] ;
              st[ic] += (*pop2).x[is].state[ic] ; }
        st[ic] /= (double)(2*M) ;
        }        
    }

    } 


/* Notice !!
   Revision date 1996/Jan/04
   A function to treat a mutual dependence is included 
*/
void initiate_pop( pop )
     population *pop ;
    {
    int is, ic ;
    double xni[STATE_DIM_MAX] ;

    /* Apr/19/1999 */
    static double  pop_vn[PARTICLE_NUM_MAX] ;

    void   get_init_pop_vn(), trans_pop2state(), trans_state2pop(), 
           init_mutdep_vn() ;

    for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
    {
       get_init_pop_vn( pop_vn, ic )    ;
       for( is = 0 ; is < M ; is++ )
             (*pop).x[is].state[ic] = pop_vn[is] ;
    }

    for( is = 0 ; is < M ; is++ )
    {
        trans_pop2state( pop, xni, is ) ;
        init_mutdep_vn( xni ) ;
        trans_state2pop( xni, pop, is ) ;
        (*pop).x[is].fitness  = 0.0  ;
    }


    } 


void copy_pop( pop1, pop2 )
     population *pop1, *pop2 ;
    {
    int is, ic ;

    for( is = 0 ; is < M ; is++ )
      {
      for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ ) {
          (*pop2).x[is].state[ic] =  (*pop1).x[is].state[ic] ; }
          (*pop2).x[is].fitness   =  (*pop1).x[is].fitness  ;
      }
    } 

/* [ resampling ] */


void resampling( fv, sel )
     double *fv    ;
     int    *sel   ;
{
     double dFinit, dF, Target, runi() ;
     int    is, ic ;

     dF = 1.0/((double)(M)) ;
     dFinit = runi()*dF ; 

     ic = 0 ;
     for( is=0 ; is < M ; is++ )
     {
         Target = ((double)(is))*dF + dFinit ;
         while( *(fv+ic) < Target ) ic++ ;

         *(sel+is) = ic ;
/*  printf(" [%4d] <%4d> %10.7f : %10.7f \n",is,ic,Target,*(fv+ic)) ; */
     }
}


void duplicate( pop, sel )
     population *pop  ;
     int    *sel      ;
    {

     /* Apr/19/1999 */
     static population tpop  ;

     int        is, ic, ip ;

    for( is = 0 ; is < M ; is++ )
      {
       for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
       {  tpop.x[is].state[ic] = (*pop).x[is].state[ic] ; }
          tpop.x[is].fitness   = (*pop).x[is].fitness   ;
      }

    for( is = 0 ; is < M ; is++ )
      {
       ip = *(sel + is) ;
       for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
       {  (*pop).x[is].state[ic] = tpop.x[ip].state[ic] ; }
          (*pop).x[is].fitness   = tpop.x[ip].fitness   ;

/*     printf(" [%4d] <%4d> %10.7f (old:new) %10.7f \n",
              is,ip,tpop.x[is].state[0],tpop.x[ip].state[0]) ; */

      }

    }


double get_dF( yn, pop, fitness )
     double *fitness, yn  ;
     population *pop  ;
     {
     double fv, ln_pop_fitness, pop_fitness, sumup, get_fitness() ;
     int    is ;

     pop_fitness = 0.0 ;
     for( is=0 ; is < M ; is++ )
     {  fv  = get_fitness( yn, pop, is ) ; pop_fitness += fv ; 
        *(fitness+is) = fv ;   }

     for( is=0 ; is < M ; is++ )
     {  *(fitness+is) /= pop_fitness ;   }
       
     sumup = (*fitness) ;
     for( is=1 ; is < M ; is++ )
     {  sumup += *(fitness+is) ;  *(fitness+is) = sumup ; }

     /* log: SING error check */
          if( pop_fitness <= MY_UNDERFLOW ) 
               ln_pop_fitness = log( MY_UNDERFLOW ) ; 
     else if( pop_fitness >= MY_OVERFLOW  ) 
               ln_pop_fitness = log( MY_OVERFLOW  ) ; 
     else ln_pop_fitness = log(pop_fitness) ;

     return( ln_pop_fitness ) ;
     }


 /* ===============================================================
                      operations of family
    =============================================================== */

void trans_family2pop( fmly, pop, lag_n )
     family  fmly[] ;
     population *pop ;
     int    lag_n ;
    {
    int is, ic ;

    for( is =0 ; is < M ; is++ )
    {
        for( ic =0 ; ic < STATE_DIM_MAX ; ic++ )
       {   (*pop).x[is].state[ic] = 
           fmly[is].history[lag_n].state[ic] ; }
    }

    }

void fixed_interval_smooth( fmly, sel, sm, fl )
     family     fmly[] ;
     int        sel[]  ;
     population *sm, *fl  ;
    {
     int        is, ic, ih, pkup ;

     /* Apr/19/1999 */
     static population tmp ;

      for( is = 0 ; is < M ; is++ )
      {
      pkup = sel[is] ;
      /* pick up particles for smoothing and filtering */
       for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
            {     (*sm).x[is].state[ic] = 
              fmly[pkup].history[(LAG_MAX-1)].state[ic] ;
                    tmp.x[is].state[ic] = 
                                (*fl).x[pkup].state[ic] ; }

              (*sm).x[is].fitness = 
          fmly[pkup].history[(LAG_MAX-1)].fitness   ;
                tmp.x[is].fitness = 
                              (*fl).x[pkup].fitness ;
      }

      /* shifting */
      for( ih = LAG_MAX-1 ; ih > 0 ; ih-- )
      {

         for( is = 0 ; is < M ; is++ )
         {    pkup = sel[is] ;
              for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
               {     fmly[is].history[ih  ].state[ic] = 
                   fmly[pkup].history[ih-1].state[ic] ; }

                     fmly[is].history[ih  ].fitness = 
                   fmly[pkup].history[ih-1].fitness ; 
          }
      }

         for( is = 0 ; is < M ; is++ )
         {    pkup = sel[is] ;
              for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
               {     fmly[is].history[0].state[ic] = 
                     (*fl).x[pkup].state[ic] ; }

                     fmly[is].history[0].fitness = 
                     (*fl).x[pkup].state[ic] ; 
          }

      for( is = 0 ; is < M ; is++ )
      {
       for( ic = 0 ; ic < STATE_DIM_MAX ; ic++ )
            { (*fl).x[is].state[ic] = tmp.x[is].state[ic] ; }
              (*fl).x[is].fitness   = tmp.x[is].fitness   ;
      }


     }



void bk_fl_smoother( y, sm, bkn, term_pop )
       double y[] ;
       population *sm, *term_pop ;
       int bkn ;
     {
    int    i, num_op, termn, 
           pop_crossover(), pop_mutation() ;
    
    /* Apr/19/1999 */
    static int sel[PARTICLE_NUM_MAX] ;
    static family fmly[PARTICLE_NUM_MAX] ;
    static double dF[PARTICLE_NUM_MAX] ;
    static population pop ;

    double ln_LK, get_dF() ;

    void   fixed_interval_smooth(), resampling(), 
           prediction(), initiate_pop(), pop_generation_shift() ;

    termn = bkn - LAG_MAX ;

    copy_pop( term_pop, &pop ) ;
    for( i=N-1 ; termn <= i ; i-- )
    {
        
        if( GA_ONOFF == 1 ) {
          pop_generation_shift( &pop ) ;
          num_op = pop_crossover( &pop ) ;
          num_op = pop_mutation( &pop ) ;
        }
        else  prediction( &pop ) ;
        
            ln_LK = get_dF( y[i], &pop, dF ) ;
            resampling( dF, sel ) ;
            fixed_interval_smooth( fmly, sel, sm, &pop ) ;
    }

    }


void fd_fl_smoother( y, sm, fdn )
       double y[] ;
       population *sm ;
       int fdn ;
     {
    int    i, num_op, termn, 
           pop_crossover(), pop_mutation() ;

    /* Apr/19/1999 */
    static int sel[PARTICLE_NUM_MAX] ;
    static double dF[PARTICLE_NUM_MAX] ;
    static family fmly[PARTICLE_NUM_MAX] ;
    static population pop ;

    double ln_LK, get_dF() ;
    void   fixed_interval_smooth(), resampling(), 
           prediction(), initiate_pop(), pop_generation_shift()  ;

    termn = fdn + LAG_MAX ;

    initiate_pop( &pop ) ;  
    for( i=0 ; i <= termn ; i++ )
    {
        
        if( GA_ONOFF == 1 ) {
          pop_generation_shift( &pop ) ;
          num_op = pop_crossover( &pop ) ;
          num_op = pop_mutation( &pop ) ;
        }
        else  prediction( &pop ) ;
        
            ln_LK = get_dF( y[i], &pop, dF ) ;
            resampling( dF, sel ) ;
            fixed_interval_smooth( fmly, sel, sm, &pop ) ;
    }

    }


void simple_twofilter_smoothing( y, sy )
       double y[], sy[] ;
     {
    int    i ;
    double st[STATE_DIM_MAX] ;

    /* Apr/19/1999 */
    static population fd_sm, bk_sm ;

    void   fd_fl_smoother(), bk_fl_smoother(), 
           get_stat_pop(), get_stat_2pop() ; 

    for( i=0 ; i < LAG_MAX; i++ )
    {
        fd_fl_smoother( y, &fd_sm, i ) ;  get_stat_pop( &fd_sm, st ) ;
        sy[ i ] = st[0]    ;
    }


    for( i=LAG_MAX ; i < N-LAG_MAX; i++ )
    {
         fd_fl_smoother( y, &fd_sm, i ) ;
         bk_fl_smoother( y, &bk_sm, i ) ;
         get_stat_2pop( &fd_sm, &bk_sm, st ) ;
         sy[ i ] = st[0]    ;
    }

    for( i=N-LAG_MAX ; i < N ; i++ )
    {
        bk_fl_smoother( y, &bk_sm, i ) ;  get_stat_pop( &bk_sm, st ) ;
        sy[ i ] = st[0]    ;
    }
    
    }


double tf_smoothing( y, sy )
       double y[], sy[] ;
     {
    int    i, num_op, lag_position, smo_position, 
           pop_crossover(), pop_mutation() ;

    /* Apr/19/1999 */
    static int  sel[PARTICLE_NUM_MAX] ;
    static double dF[PARTICLE_NUM_MAX] ;
    static family fmly[PARTICLE_NUM_MAX] ;
    static population pop, fd_sm, bk_sm, term_pop ;

    double tplk, ln_LK, get_dF(), st[STATE_DIM_MAX], 
           filtering() ;
    void   fixed_interval_smooth(), resampling(), get_stat_pop(), 
           prediction(), trans_family2pop(), pop_generation_shift() ;


    initiate_pop( &pop ) ;  
    copy_pop( &pop, &term_pop ) ;

   /* get a final PDF */
    for( i=0 ; i < N; i++ )
    {
    if( GA_ONOFF == 1 ) {
       pop_generation_shift( &term_pop ) ;
       num_op = pop_crossover( &term_pop ) ;
       num_op = pop_mutation( &term_pop ) ;
     }
    else  prediction( &term_pop ) ;
    tplk = filtering( y[i], &term_pop ) ;
    }


    for( i=0 ; i < N; i++ )
    {

    if( GA_ONOFF == 1 ) {
      pop_generation_shift( &pop ) ;
      num_op = pop_crossover( &pop ) ;
      num_op = pop_mutation( &pop ) ;
    }
    else  prediction( &pop ) ;

    ln_LK = get_dF( y[i], &pop, dF ) ;  tplk += ln_LK ;
    resampling( dF, sel )   ;
    fixed_interval_smooth( fmly, sel, &fd_sm, &pop ) ;

    smo_position = i-LAG_MAX ;
    if( smo_position >= LAG_MAX ) {
         bk_fl_smoother( y, &bk_sm, smo_position, &term_pop ) ;
            if( PrintM == i-LAG_MAX ) print_pop( &fd_sm ) ;
            if( PrintM == i-LAG_MAX ) print_pop( &bk_sm ) ;
         get_stat_2pop( &fd_sm, &bk_sm, st ) ;
         sy[ smo_position ] = st[0]    ; }
    else if( smo_position >= 0 ) {
         get_stat_pop( &fd_sm, st ) ; 
         sy[ smo_position ] = st[0]      ; }

    }


    for( i=N-LAG_MAX ; i < N ; i++ )
    {
        lag_position = i-(N-LAG_MAX) ; 
        trans_family2pop( fmly, &fd_sm, lag_position ) ; 

        bk_fl_smoother( y, &bk_sm, i, &term_pop ) ;
        get_stat_2pop( &fd_sm, &bk_sm, st ) ;
        sy[ i ] = st[0]    ;
    }

    return( tplk ) ;
    }

/* Version UP 1995/11/28  */
/* Version UP 1996/01/22  */
/* Version UP 1996/01/29  */
double smoothing( y, sy )
       double y[], sy[][DATA_MAX] ;
     {
    int    i, ic, num_op, lag_position, data_position, 
           pop_crossover(), pop_mutation() ;

    /* Apr/19/1999 */
    static int sel[PARTICLE_NUM_MAX] ;
    static double dF[PARTICLE_NUM_MAX] ;
    static family fmly[PARTICLE_NUM_MAX] ;
    static population pop, sm ;

    double tplk, ln_LK, get_dF(), st[(ULout*SHOW_MAX)] ; 
    void   fixed_interval_smooth(), resampling(), get_stat_pop(), 
           prediction(), trans_family2pop(), pop_generation_shift() ;

    tplk = 0.0 ;
    initiate_pop( &pop ) ;  

    for( i=0 ; i < N; i++ )
    {
    LOC_n = i ;

    if( GA_ONOFF == 1 ) {
      pop_generation_shift( &pop ) ;
      num_op = pop_crossover( &pop ) ;
      num_op = pop_mutation( &pop ) ;
    }
    else  prediction( &pop ) ;

    if( PrintM == i ) print_pop( &pop ) ;

    if( y[i] != MISSING ) {
        ln_LK = get_dF( y[i], &pop, dF ) ;  tplk += ln_LK ;
        resampling( dF, sel )   ;
    }

    fixed_interval_smooth( fmly, sel, &sm, &pop ) ;

    if( PrintM == i ) print_pop( &pop ) ;

    get_stat_pop( &sm, st ) ;
    if( i >= LAG_MAX ) { 
          for( ic=0 ; ic < (ULout*SHOW_MAX) ; ic ++ ) 
                        sy[ic][i-LAG_MAX] = st[ic] ;}
    }

    for( i=0 ; i < LAG_MAX ; i++ )
    {  
       lag_position = LAG_MAX-1-i ; data_position = N-LAG_MAX+ i ;
       trans_family2pop( fmly, &sm, lag_position ) ; 
       get_stat_pop( &sm, st ) ;
       for( ic=0 ; ic < (ULout*SHOW_MAX) ; ic ++ ) 
                       sy[ic][data_position] = st[ic] ;
    }

    return( tplk ) ;
    }


/* Version UP 1995/11/28  */
double parallel_smoothing( y, sy )
       double y[], sy[][DATA_MAX] ;
     {
     
      /* Apr/19/1999 */
      static double smlnlk[SMOOTH_PAL_MAX], 
                    tmp[(ULout*SHOW_MAX)][DATA_MAX] ;

      double smoothing(), tpv, lk_max ;
      int ip, is, ic, n_max ;

      lk_max = -MY_OVERFLOW ; n_max = 1 ;
      tpv = 0.0 ;
      for( ic=0 ; ic < (ULout*SHOW_MAX) ; ic++ ) {
         for( is=0 ; is < N ; is++ ) sy[ic][is] = 0.0 ;  }

      for( ip=0 ; ip < SMOOTH_PAL_MAX ; ip++ ) 
      {    
           smlnlk[ip] = smoothing( y, tmp ) ;

           if( PAL_SMO_AVE == TRUE )
	   {
              for( ic=0 ; ic < (ULout*SHOW_MAX) ; ic++ ) 
                   for( is=0 ; is < N ; is++ ) sy[ic][is] +=  tmp[ic][is]; 
              tpv += smlnlk[ip] ; 
           }
           else if( smlnlk[ip] >= lk_max ) 
           {
              for( ic=0 ; ic < (ULout*SHOW_MAX) ; ic++ ) 
                   for( is=0 ; is < N ; is++ ) sy[ic][is]  =  tmp[ic][is]; 
              tpv  = smlnlk[ip] ; 
           }

           if( smlnlk[ip] >= lk_max ) 
           {   lk_max = smlnlk[ip] ;  n_max = ip+1 ;  
               if( PAL_SMO_AVE == TRUE )  tpv += smlnlk[ip] ; 
               else                       tpv  = smlnlk[ip] ; 
           }

           printf(" [%4d] <AIC> = %10.5f  AIC_min %10.5f \n", ip+1, 
                  -2*smlnlk[ip], -2*lk_max) ;
      }

      if( PAL_SMO_AVE == TRUE ) 
      {
         for( ic=0 ; ic < (ULout*SHOW_MAX) ; ic++ ) {
            for( is=0 ; is < N ; is++ ) 
                  sy[ic][is] /= (double)(SMOOTH_PAL_MAX) ; }
         tpv /= (double)(SMOOTH_PAL_MAX) ;
      }

      tpv = -2.0*tpv ; 

      printf(" <AIC>_%4d = %10.5f AIC_min[%4d] %10.5f \n",SMOOTH_PAL_MAX,
              tpv, n_max, -2*lk_max) ;

      return( tpv ) ;

     }

 /* ===============================================================
                        output routine 
    =============================================================== */

double parallel_get_lnYN( y, stat_lk )
       double y[], stat_lk[] ;
    {
    double lkv[PRDFIL_PAL_MAX], ave_lk, get_lnYN() ;

    /* Apr/19/1999 */
    static population pop ;

    int jj, med, upq, lowq ;
    void sort() ;

    ave_lk = 0.0 ;    initiate_pop( &pop )  ;

    for( jj=0 ; jj < PRDFIL_PAL_MAX ; jj++)
   {
    lkv[jj] = get_lnYN( y ) ; ave_lk += lkv[jj] ;
    }
    ave_lk /= (double)(PRDFIL_PAL_MAX) ;

    sort( PRDFIL_PAL_MAX, lkv ) ;

    lowq = (int)( (double)(PRDFIL_PAL_MAX)*0.25 ) ;
     med = (int)( (double)(PRDFIL_PAL_MAX)*0.50 ) ;
     upq = (int)( (double)(PRDFIL_PAL_MAX)*0.75 ) ;

    stat_lk[0] = lkv[0]    ;
    stat_lk[1] = lkv[lowq] ;
    stat_lk[2] = lkv[med ] ;
    stat_lk[3] = lkv[upq ] ;
    stat_lk[4] = lkv[PRDFIL_PAL_MAX-1] ;

    return( ave_lk ) ;
    }


/* indexes an array "arrin[1..n]" outputs the array "indx[1..n]" 
such that "arrin[ indx[j] ]" os in ascending order for "j=1,2,.. N". */

void indexx(n, y, ix)
int n, ix[] ;
double y[] ;
{

     int l, j, ir, i, indxt ;
     
     /* Apr/19/1999 */
     static int indx[(PARTICLE_NUM_MAX+1)] ;
     static double  arrin[(PARTICLE_NUM_MAX+1)] ;

     double q  ;

     for ( j=1 ; j <=n; j++ ) {
       indx[j]=j ; arrin[j] = y[j-1] ;
     }
     if( n==1 ) return ;
     l = ( n >> 1 ) + 1 ;
     ir = n ; 
     for ( ;; ) {
          if ( l > 1 )
                q=arrin[(indxt=indx[--l])] ;
          else {
                q=arrin[(indxt=indx[ ir])] ;
                indx[ir] = indx[1] ;
                if( --ir == 1 ) {
                     indx[1] = indxt ;
                     goto sortend ;
                }
          }
          i = l ;
          j = l << 1 ;
          while( j <= ir ) {
                if( j < ir && arrin[indx[j]] < arrin[indx[j+1]]) j++ ;
                if( q < arrin[indx[j]]) {
                     indx[i] = indx[j] ;
                     j += (i=j) ;
                }
                else  j = ir+1 ;
          }
          indx[i] = indxt ;
    }
     sortend :
     for ( j=1 ; j <=n; j++ ) {
         ix[j-1]=indx[j]-1 ;
     }

}

void rand_sort( ns, ix )
int ns, ix[] ;
{
    double runi() ;

    /* Apr/19/1999 */
    static double ru[PARTICLE_NUM_MAX] ;

    int    i ;
    void indexx() ;

    for( i=0 ; i < ns ; i++) ru[i] = runi() ;
    indexx(ns, ru, ix) ;
/*
    for( i=0 ; i < ns ; i++) 
    printf(" [%3d] %7.5f \n",i,ru[ ix[i] ]) ;
*/
}

void sort(n, ru)
int n ;
double ru[] ;
{
     int l, j, ir, i ;

     /* Apr/19/1999 */
     static double ra[(2*PARTICLE_NUM_MAX+1)] ;

     double rra ;

     for( i=0 ; i < n ; i++) ra[i+1] = ru[i] ;

     l = ( n >> 1 )+1 ;
     ir = n ;

     for (;;) {
         if( l > 1 ) 
              rra = ra[--l] ;
         else {
              rra = ra[ir]  ;
              ra[ir] = ra[1] ;
              if( --ir == 1 ) {
                    ra[1] = rra ;
                    goto sortend ;
              }
         }
         i = l ;
         j = l << 1 ;
         while( j <= ir ) {
              if( j < ir && ra[j] < ra[j+1] ) ++j ;
              if( rra < ra[j] ) {
                  ra[i] = ra[j] ;
                     j += (i=j) ;
              }
              else   j = ir + 1 ;
         }
         ra[i] = rra ;
     }

     sortend :
     for( i=0 ; i < n ; i++) ru[i] = ra[i+1] ;

 }


 /* ===============================================================
                       system definition 
    =============================================================== */

void define_mcf_system() 

{
    K = STATE_DIM_MAX    ;  M = PARTICLE_NUM_MAX ;
    L = SYSNOISE_DIM_MAX ;

    xmin_gpdf =  -(double)( Gauss_dF_range ) ;
    xmax_gpdf =   (double)( Gauss_dF_range ) ;
    dx_gpdf   =  (xmax_gpdf - xmin_gpdf)/(double)(Gauss_dF_DIM_MAX) ;
    dx_unif   =  (1.0  - 0.0 )/(double)(Gauss_dF_DIM_MAX) ;

}

