

#include <math.h>
#include <stdio.h>

static  int  Icount = 0 ;
static double S_rgauss, V2_rgauss ;


/* ===============================================================
                      definition of struct
    ============================================================== */

/* particle */
struct stat_of_data {
       int    num, num_real ;
       double min, max, mean, var ;
     } ;
typedef struct stat_of_data STAT ;
static  STAT stat_y ;

 /* ===============================================================
                           noise form 
    =============================================================== */

/* uniform distribution */

 double  uniform(b_value, mean, tau2, x)

        double b_value, mean, tau2, x;
       {
        return( 1.0 );
        }

/*  gaussian distrubution */
 double  gauss(b_value, mean, sig2, x)

        double b_value, mean, sig2, x;
       {
        double   pf, denom, x2 ;

        x -= mean ; x2 = x*x ;
        denom = 1/( 2.0 * sig2 );
        pf = sqrt( denom/ PI ) * exp( -denom*x2 ) ;
        return( pf );
        }

/* cauchy distribution */

double   cauchy(b_value, mean, tau2, x)

        double b_value, mean, tau2, x;
       {
        double   pf, tau, x2;
        tau = sqrt( tau2 ) ;  x -= mean ;
        x2 = x*x  ;
        pf = tau/( PI *( tau*tau + x2 )) ;
        return( pf );
        }

/* gamma function provies a value of log-gamma */
double gammaln(xx)
double xx ;
{
      double x, tmp, ser ;
      static double cof[6]={76.18009173, -86.50532033, 24.01409822,
      -1.231739516, 0.120858003e-2, -0.536382e-5 };
      int j ;

      x=xx-1.0 ;
      tmp=x+5.5 ;
      tmp -= (x+0.5)*log(tmp) ;
      ser = 1.0 ;
      for(j=0; j<=5; j++) {
      x += 1.0 ;
      ser += cof[j]/x ;
      }
      return( -tmp+log(2.50662827465*ser)) ;
}

/* pearson family distribution */
double pearson(b_value, mean, tau2, x)
        double b_value, mean, tau2, x;
       {
        double   c_value, denom, pf, tau, x2 ;
        double    gammaln() ;

        tau = sqrt( tau2 ) ;  x -= mean ;    x2 = x*x  ;

        denom = -b_value * log(tau*tau + x2) ;
        c_value = (2.0*b_value-1)*log( tau ) + gammaln( b_value ) -
        gammaln(0.5) - gammaln( b_value-0.5 ) ;
        pf = c_value + denom ;
        return( (exp(pf)) );
        }

 /* ===============================================================
                    random noise generation 
    =============================================================== */
/* uniform (1) Normalized */
double runi()
      {
       return( (double)(rand())/(double)(RAND_MAX) ) ;
       }

/* uniform (2) Not Normalized */
double runiform(b_value, min_x, max_x)
         double b_value, min_x, max_x ;
      {
       double tmp ;
       tmp = rand()/RAND_MAX + 1.0 ;
       tmp = (max_x - min_x)*runi() + min_x ;
       return( tmp ) ;
       }

/* Gauss */
double rgauss(b_value, mean, var )
       double b_value, mean, var ;
      {
      double tmp, U1, U2, V1 ;

      if( Icount == 0 )
      {  do
           {
             U1 = runi() ; U2 = runi() ;
             V1 = 2.0*U1 - 1.0 ; V2_rgauss = 2.0*U2 - 1.0 ;
             S_rgauss = V1*V1 + V2_rgauss*V2_rgauss ;
            }
         while( S_rgauss >= 1.0 ) ;
             tmp = V1*sqrt( -2.0*log(S_rgauss)/S_rgauss ) ;
          Icount = 1 ;
       }
      else
      {  
             tmp = V2_rgauss*sqrt( -2.0*log(S_rgauss)/S_rgauss ) ; 
          Icount = 0 ;
       }

       tmp = sqrt( var )*tmp + mean ;
       return( tmp ) ;

       }


/* Cauchy */
double rcauchy(b_value, mean, var )
       double b_value, mean, var ;
      {
       double tmp, U ;

       do 
       {
           do U = runi() ; while( U == 0.5 ) ; 
           tmp = tan(PI*U ) ;
           tmp = sqrt(var)*tmp + mean ;
       }
       while( tmp < R_MIN || tmp > R_MAX ) ;

       return( tmp ) ;
       }


void stochastic_uniformal( nd, vn, vmin, vmax )
     double *vn, vmin, vmax ;
     int     nd ;
   {
     double runi(), dF, dFi ;

     /* Apr/19/1999 */
     static double  un[PARTICLE_NUM_MAX] ;
     static int ix[PARTICLE_NUM_MAX] ;

     int    is ;
     void   rand_sort() ;

     rand_sort(M, ix) ;
     dF  = (vmax - vmin)/((double)(M)) ;
     dFi =  vmin + dF*runi() ;

     for( is=0 ; is < M ; is++ )
     un[is] =  dFi + (double)(is)*dF ; 

     for( is=0 ; is < M ; is++ )
     *( vn+is ) = un[ ix[is] ] ;
   }

void dF_gauss()
   {
    int i, is, ic ;
    double x1, x2, f1, f2, pdf_norm, sum, Target, 
           gauss(), pdf[Gauss_dF_DIM_MAX], get_x_pdf_gauss() ;

    for( i=0 ; i < Gauss_dF_DIM_MAX ; i++ ) {
       x1 = xmin_gpdf + (double)(i  )*dx_gpdf ;
       x2 = xmin_gpdf + (double)(i+1)*dx_gpdf ;

       f1 = gauss(0.0, 0.0, 1.0, x1) ;
       f2 = gauss(0.0, 0.0, 1.0, x2) ;

       pdf[i] = 0.5*(f1+f2)*dx_gpdf ;
    }

    pdf_norm = 0.0 ;
    for( i=0 ; i < Gauss_dF_DIM_MAX ; i++ ) pdf_norm += pdf[i] ;
    
    for( i=0 ; i < Gauss_dF_DIM_MAX ; i++ ) 
    pdf[i] /= pdf_norm ;

    sum = 0.0 ;
    for( i=0 ; i < Gauss_dF_DIM_MAX ; i++ ) {
    sum += pdf[i] ; pdf[i] = sum ; }

    ic      = 0 ;
                       dF[0] = -(double)(Gauss_dF_range) ;
    dF[(Gauss_dF_DIM_MAX-1)] =  (double)(Gauss_dF_range) ;

/*    printf(" %10.7f \n",dF[0]) ; */
    for( is=1 ; is < (Gauss_dF_DIM_MAX-1) ; is++ )
    {
        Target = ((double)(is+1))*dx_unif ;
        while( pdf[ic] < Target ) ic++ ;

        dF[is] = get_x_pdf_gauss( ic ) ;
/*    printf(" %10.7f \n",dF[is]) ; */
    }
/*    printf(" %10.7f \n",dF[(Gauss_dF_DIM_MAX-1)]) ; */

   }

double get_x_pdf_gauss( is )
int is ;
    {
double x1, x2 ;

    x1 = xmin_gpdf + (double)(is  )*dx_gpdf ;
    x2 = xmin_gpdf + (double)(is+1)*dx_gpdf ;
    return( (x1 + x2)*0.5 ) ;
    }


int get_pnt_pdf_gauss( vn )
double vn ;
    {
int rtin  ;

    rtin = (int)( (vn/dx_unif) ) ;
    return( rtin ) ;
    }

void stounif_gauss( nd, vn, mu, tau2)
     double vn[], mu, tau2 ;
     int      nd ;
   {
    int     i, isel, get_pnt_pdf_gauss() ;

    /* Apr/19/1999 */
    static double  un[PARTICLE_NUM_MAX] ;

    double  tmp, U, x1, x2 ;
    void stochastic_uniformal() ;

    stochastic_uniformal( nd, un, 0.0, 1.0 ) ;

    for( i=0 ; i < nd ; i++ ) {
        U = un[i] ; isel = get_pnt_pdf_gauss( U ) ;
        x1 = dF[ isel ] ; 
        if( isel == Gauss_dF_DIM_MAX-1 ) {
        x2 = dF[ isel-1 ] ; tmp  = x2 + runi()*(x1-x2) ; }
        else {
        x2 = dF[ isel+1 ] ; tmp  = x1 + runi()*(x2-x1) ; }
        tmp *= sqrt(tau2) ; vn[i] = tmp + mu ;
/*      printf(" [%5d] isel= %5d %12.8f %12.8f \n",i,isel,U,vn[i]) ;*/
    }


    }


void stounif_cauchy( nd, vn, mu, tau2 )
     double vn[], mu, tau2 ;
     int      nd ;
   {
     int     i, np ;

     /* Apr/19/1999 */
     static double un[PARTICLE_NUM_MAX] ;

     double  U, umin, umax, tmp ;
     void stochastic_uniformal() ;
     
     /* 2002. Nov. 20 */
     umin = (double)(R_MIN) ; umax = (double)(R_MAX) ; 
     umin /= sqrt(tau2) ; umax /= sqrt(tau2) ; 
     umin = atan( umin )/PI + 0.5 ;
     umax = atan( umax )/PI + 0.5 ;

/*     printf(" x: %10.5f -- %10.5f   u: %10.5f -- %10.5f \n",
     R_MIN,R_MAX,umin,umax) ; */

     stochastic_uniformal( nd, un, umin, umax ) ;

     for( i=0 ; i < nd ; i++ ) {
        U = un[i]-0.5 ;
        if( U ==  0.5 ) U -= MY_UNDERFLOW ;
        if( U == -0.5 ) U += MY_UNDERFLOW ;
     
        tmp = tan( PI*U ) ;
        tmp = sqrt(tau2)*tmp + mu ;
        vn[i] = tmp ; 
     }

   }

void get_vn( vn, iv )
     double *vn ;
     int iv ;
/*  V-UP 1995/11/28 */
    {
     double bv, mu, tau2 ;

    bv = sys_type[iv][0] ; mu = sys_type[iv][1] ; tau2 = sys_type[iv][2] ; 

         if( bv == 0.0 ) *vn =   rgauss(bv, mu, tau2) ; 
    else if( bv == 1.0 ) *vn =  rcauchy(bv, mu, tau2) ; 
    else                 *vn = runiform(bv, mu, tau2) ;

    }


void get_init_vn( vn, ic )
     double *vn ;
     int     ic ;
    {
     double bv, mu, tau2 ;

    bv = init_type[ic][0] ; mu = init_type[ic][1] ; tau2 = init_type[ic][2] ; 

         if( bv == 0.0 ) *vn =   rgauss(bv, mu, tau2) ; 
    else if( bv == 1.0 ) *vn =  rcauchy(bv, mu, tau2) ; 
    else                 *vn = runiform(bv, mu, tau2) ;

    }

void get_pop_vn( pop_vn, iv )
     double pop_vn[] ;
     int    iv       ;
/*  Version UP 1995/11/18  for analyzing the rain.org 
    "iv" was input in a get_pop_vn(( )  */
   {
     double  vn[SYSNOISE_DIM_MAX], ur ;
     int     is ;
     void   get_vn(), stochastic_uniformal(), stounif_gauss(), 
            stounif_cauchy() ;


     if( sys_type[iv][0] == 0.0 ) 
         stounif_gauss( M, pop_vn, sys_type[iv][1], sys_type[iv][2]) ;
     else if( sys_type[iv][0] == 1.0 ) 
        stounif_cauchy( M, pop_vn, sys_type[iv][1], sys_type[iv][2]) ;
     else {
        ur = sys_type[iv][2] - sys_type[iv][1] ;
        stochastic_uniformal( M, pop_vn, 0.0, 1.0 ) ;
               for( is=0 ; is < M ; is++ )
               pop_vn[is]  =  sys_type[iv][1] + ur*pop_vn[is];
          }

/*
          {
               for( is=0 ; is < M ; is++ )
              { get_vn( vn, iv ) ;   *( pop_vn+is ) = vn[0] ; }
          }
*/
   }


void get_init_pop_vn( pop_vn, ic )
     double pop_vn[] ;
     int    ic       ;
   {
     double  vn[SYSNOISE_DIM_MAX], ur ;
     int     is ;
     void   get_init_vn(), stounif_gauss(), stounif_cauchy(), 
            stochastic_uniformal() ;

     if( init_type[ic][0] == 0.0 ) 
         stounif_gauss( M, pop_vn, init_type[ic][1], init_type[ic][2]) ;
     else if( init_type[ic][0] == 1.0 ) 
        stounif_cauchy( M, pop_vn, init_type[ic][1], init_type[ic][2]) ;
     else {
        ur = init_type[ic][2] - init_type[ic][1] ;
        stochastic_uniformal( M, pop_vn, 0.0, 1.0 ) ;
               for( is=0 ; is < M ; is++ )
              pop_vn[is]  =  init_type[ic][1] + ur*pop_vn[is];
          }

/*      {
        for( is=0 ; is < M ; is++ )
        { get_init_vn( vn, ic ) ;   *( pop_vn+is ) = vn[0] ; }
        }
*/
   }

void prediction( pop )
     population *pop ;
    {
     double vn[SYSNOISE_DIM_MAX], xn[STATE_DIM_MAX] ;

     /* Apr/19/1999 */
     static double pop_vn[PARTICLE_NUM_MAX], 
            multi_pop_vn[SYSNOISE_DIM_MAX][PARTICLE_NUM_MAX] ;

     int is, iv ;
     void get_pop_vn(), one_ahead() ;

     for( iv = 0 ; iv < SYSNOISE_DIM_MAX ; iv++ )
        {
          get_pop_vn( pop_vn, iv ) ;
          for ( is=0 ; is < M ; is++ )
             multi_pop_vn[iv][is] = pop_vn[is] ;
        }


     for( is=0 ; is < M ; is++ )
     {   
         for( iv = 0 ; iv < SYSNOISE_DIM_MAX ; iv++ )
            vn[iv] = multi_pop_vn[iv][is] ;
 
         trans_pop2state( pop, xn, is ) ;
         one_ahead( xn, vn ) ;
/*       printf("%10.5f \n",vn[0]) ; */
         trans_state2pop( xn, pop, is ) ; }
    }

/*  [ for filtering ] */

double filtering( yn, pop )
     double yn ;
     population *pop ;
    {
     double  ln_LK, get_dF() ;

     /* Apr/19/1999 */
     static double dF[PARTICLE_NUM_MAX] ;
     static int    sel[PARTICLE_NUM_MAX] ; 

     void   resampling(), duplicate() ;

     ln_LK = get_dF( yn, pop, dF ) ; 
     resampling( dF, sel ) ;  duplicate( pop, sel ) ;

     return( ln_LK ) ;
    }


double get_fitness( yn, pop, particle_num )
    population *pop ;
    double      yn  ;
    {
    double get_likelihood(), fv, zn[STATE_DIM_MAX] ;

    trans_pop2state( pop, zn, particle_num ) ;
    fv  = get_likelihood( yn, zn ) ; 
    fv /= ((double)(M)) ;
    (*pop).x[particle_num].fitness = fv ;

    return( fv ) ;
    }


double get_lnYN( y )
       double *y ;
     {
    
    /* Apr/19/1999 */
    static population pop ;

    double tplk, tmp, filtering() ;
    int    i, num_op, pop_crossover(), pop_mutation() ;
    void prediction(), initiate_pop(), pop_generation_shift()  ;

    tplk = 0.0 ;

    initiate_pop( &pop ) ;  
    for( i=0 ; i < N; i++ )
    {
    if( GA_ONOFF == 1 ) {
       pop_generation_shift( &pop ) ;
       num_op = pop_crossover( &pop ) ;
       num_op = pop_mutation( &pop ) ;
     }
    else  prediction( &pop ) ;

    tmp = filtering( *(y+i), &pop ) ;
    tplk += tmp ;
    }

    return( tplk ) ;
    }

/*  [ for print ] */

void print_state( x )
     double *x ;
    {
    int i ;
    for( i =0 ; i < STATE_DIM_MAX ; i++ )
    {  printf(" [%1d] %10.4f", i, *(x+i)) ;}
       printf("\n") ;
    }  

void define_state( x )
     double *x ;
    {
       *(x+0) = 0.1 ;
       *(x+1) = 0.0 ;
    }  



 /* ===============================================================
                         input data routine 
              F means double precision   -+d.ddd 
              E                          -+d.ddd E-+eee 
              f and e: single precision   
              lf  treats both F & E

    =============================================================== */

/*   read data function   */
void read_data(file_name, observation )
     double  *observation ;
     char      *file_name ;
     {
      int            num, tp_v  ;
      FILE           *fp_r ;
      void stat_of_time_series() ;

      /* --- file open --- */

         if(NULL == (fp_r = fopen(file_name, "r")))
          {
           printf(" can not open file  : %s \n", file_name) ;
           /* file does not exist */
           exit(0);
          }

      /* --- reading --- */
        num = 0 ;
        while( EXIST_FILE == feof(fp_r) )
        {  /* while (2) start */
           num ++ ;
           if( Y_TYPE_INT == ON ) 
           {
              fscanf(fp_r,"%d \n", &tp_v) ;
              *(observation+num-1) = (double)tp_v ;
           }
           else
           {
              fscanf(fp_r,"%lf \n", (observation+num-1)) ;
           }

           if( LOG_Y == 1 && *(observation+num-1) != MISSING )  
           *(observation+num-1) = log10( *(observation+num-1) ) ;

        }  /* while (2) stop  */

/*  --- file close --- */
      N = num ;
      fclose(fp_r) ;
      stat_of_time_series( observation ) ;

      }


void stat_of_time_series( y ) 
     double *y ;

     {
      int   i, num_real ;
      double  temp_mean, temp_var, temp_min, temp_max ; 


      if  ( *(y) != MISSING )
      {
         temp_min  = *(y+0) ; temp_max  = *(y+0) ;
         temp_mean = *(y+0) ; num_real  = 1 ;
      }
      else
      {
         temp_min  =  MY_OVERFLOW ; temp_max  =  MY_UNDERFLOW ;
         temp_mean =  NA          ; num_real  = 0 ;
      }


      for( i=1 ; i < N ; i++ )
      {
        if( ( *(y+i) < temp_min ) && ( *(y+i) != MISSING ) )
                temp_min = *(y+i) ;

        if( ( temp_max  < *(y+i)) && ( *(y+i) != MISSING ) )
                temp_max = *(y+i) ;
        if  ( *(y+i) != MISSING )
           {  num_real++ ; temp_mean += *(y+i) ; }
      }
      temp_mean /= (double)(num_real) ;

      temp_var = 0.0 ;
      for( i=0 ; i < N ; i++ )
      {
        if  ( *(y+i) != MISSING )
        temp_var += (*(y+i)-temp_mean)*(*(y+i)-temp_mean) ;
      }
      if( N < 2 )
       { printf("\7***** variance can not defined (N<2) ****\n"); }
      else
       { temp_var /= (double)(num_real - 1) ;}

      
      stat_y.num  = N         ; stat_y.num_real = num_real ;
      stat_y.min  = temp_min  ; stat_y.max = temp_max ;
      stat_y.mean = temp_mean ; stat_y.var = temp_var ;

      printf("    \n") ;
      printf("    \n") ;
      printf(" -------------- simple statistics of data -------------  \n") ;
      printf("    \n") ;
      printf(" total data number     : %6d \n", stat_y.num) ;
      printf(" number of missing data: %6d \n", 
             stat_y.num - stat_y.num_real ) ;
      printf("    \n") ;
      printf(" mean & variance       : %10.5f   and   %10.5f \n",
             stat_y.mean, stat_y.var ) ;
      printf(" data range            : %10.5f  <--->  %10.5f \n", 
             stat_y.min, stat_y.max ) ;
      printf(" ------------------------------------------------------  \n") ;
      printf("    \n") ;

      }



 /* ===============================================================
                        output routine 
    =============================================================== */
/* Version UP 1995/11/28  */
/*   output of time series in ASCII format */
     void time_series_ascii_output( y, result, file_name )
     double  result[][DATA_MAX], y[] ;
     char   *file_name ;
    {
     int i, ic ;
     double yv ;
     FILE   *fp_w ;


      /* --- file open --- */

         if(NULL == (fp_w = fopen(file_name, "w")))
          {
           printf(" can not open file  : %s \n", file_name) ;
           /* file does not exist */
           exit(0);
          }

     for( i = 0; i < N ; i++ )
          {
              fprintf(fp_w, "%f ", y[i]);
              for( ic=0 ; ic < (ULout*SHOW_MAX)-1 ; ic++ ) 
              fprintf(fp_w, "%f ", result[ic][i]) ;
              fprintf(fp_w, "%f \n", result[(ULout*SHOW_MAX-1)][i]) ;
	  }
     fclose( fp_w ) ;
     }

